/***************************************************************************
 *
 * Copyright (c) 2000, 2001 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: tisauth.h,v 1.3.2.2 2001/09/19 19:22:17 bazsi Exp $
 *
 ***************************************************************************/
#ifndef _TISAUTH_H_INCLUDED
#define _TISAUTH_H_INCLUDED

#include <zorp/zorp.h>
#include <zorp/sockaddr.h>

typedef struct _ZTISContext ZTISContext;

ZTISContext *z_tis_new(ZSockAddr *addr);
gboolean z_tis_connect(ZTISContext *ctx);
void z_tis_destroy(ZTISContext *ctx);
gboolean z_tis_authenticate(ZTISContext *ctx, gchar *challenge, gint challenge_size, gchar *user);
int z_tis_response(ZTISContext *ctx, char *response);
char *z_tis_fake_challenge(void);
gboolean z_tis_challenge(ZTISContext *ctx, gchar *challenge, gint challenge_size, gchar *user);
int  z_tis_respchal(ZTISContext *ctx, char *username, char *challenge, char *response);

#endif

